/*--------------------------------------------------------------------------+
|                                                                          |
|                                                                          |
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.executable;

import static org.fortiss.af3.expression.utils.ExpressionModelElementFactory.funcCall;
import static org.fortiss.af3.expression.utils.ExpressionModelElementFactory.rawString;

import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.expression.model.terms.IExpressionTerm;
import org.fortiss.af3.expression.model.terms.RawString;
import org.fortiss.af3.expression.model.types.TBool;
import org.fortiss.af3.expression.model.types.TDouble;
import org.fortiss.af3.expression.model.types.TInt;
import org.fortiss.af3.platform.language.executable.TransceiverExecutableBase;
import org.fortiss.af3.platform.raspberry.model.CanConnector;

/**
 * Executable for {@link CanConnector}.
 * 
 * @author eder
 * @author  hoelzl $
 *                                                                                                 18709 $
 *                                                                                                 RED Hash:
 */
public class CanConnectorExecutable extends TransceiverExecutableBase<CanConnector> {

	/** Constructor. */
	public CanConnectorExecutable(CanConnector modelElement) {
		super(modelElement);
	}

	public IExpressionTerm getInitialization() {
		// init CAN bus on the given interface
		return funcCall("can_init");
	}

	public IExpressionTerm getNoValGuardAccessor(InputPort logicalSignal) {
		// check for NOVAL on the given signal port
		return funcCall("can_is_noval", literalString(logicalSignal.getName()));
	}

	public IExpressionTerm getNoValWriteAccessor(OutputPort logicalSignal) {
		// set NOVAL on the given signal port
		return funcCall("can_set_noval", literalString(logicalSignal.getName()));
	}

	public IExpressionTerm getTermination() {
		// terminate the bus on the given interface
		return funcCall("can_term");
	}

	/** {@inheritDoc} */
	@Override
	public IExpressionTerm getValueReadAccessor(InputPort logicalSignal) {
		// read the value on the given signal port depending on the variable type
		if(logicalSignal.getVariableType() instanceof TBool) {
			return funcCall("can_read_bool", literalString(logicalSignal.getName()));
		} else if(logicalSignal.getVariableType() instanceof TDouble) {
			return funcCall("can_read_double", literalString(logicalSignal.getName()));
		} else if(logicalSignal.getVariableType() instanceof TInt) {
			return funcCall("can_read_int", literalString(logicalSignal.getName()));
		} else {
			throw new RuntimeException("unknown data type: " +
					logicalSignal.getVariableType().getTypeClassName());
		}
	}

	/** {@inheritDoc} */
	@Override
	public IExpressionTerm getValueWriteAccessor(OutputPort logicalSignal, IExpressionTerm value) {
		// write the value on the given signal port depending on the variable type
		if(logicalSignal.getVariableType() instanceof TBool) {
			return funcCall("can_write_bool", literalString(logicalSignal.getName()), value);
		} else if(logicalSignal.getVariableType() instanceof TDouble) {
			return funcCall("can_write_double", literalString(logicalSignal.getName()), value);
		} else if(logicalSignal.getVariableType() instanceof TInt) {
			return funcCall("can_write_int", literalString(logicalSignal.getName()), value);
		} else {
			throw new RuntimeException("unknown data type: " +
					logicalSignal.getVariableType().getTypeClassName());
		}
	}

	/**
	 * builds a constant literal C string including double quotes
	 */
	private RawString literalString(String str) {
		return rawString("\"" + str + "\"");
	}
}
